/*
 * =====================================================================================
 *
 *       Filename:  lynx_filelist.cc
 *
 *    Description:  lynx_filelist recursively walks through a directory and generates a
 *                  list of all files having specific extensions.
 *
 *        Version:  1.0
 *        Created:  10/26/2007 08:58:29 PM CEST
 *       Revision:  01/07/2007 16:30:00 PM CET
 *       Compiler:  g++
 *
 *         Author:  R. Pfeiffer (Mr), lynx@luchs.at
 *        Company:  http://web.luchs.at/
 *
 * =====================================================================================
 */

//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

//-----------------------------------------------------------------------------
// Include section
//-----------------------------------------------------------------------------

#include <iostream>
#include <iomanip>
#include <string>
#include <fstream>
#include <exception>
#include <locale>
#include <vector>

#include <time.h>

#include <libconfig.h++>
#include <boost/lexical_cast.hpp>

#include "cmdline.h"
// Please note that we share code with the file helper.cc. The
// walkthrough() function is part of this file and it does the main
// work of our program (i.e. walking through a tree in the filesystem
// and taking notes).
#include "helper.h"

using namespace libconfig;
using namespace std;

//-----------------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------------  

// Structure with command line options
gengetopt_args_info args_info;

//----------------------------------------------------------------------
// Main section
//----------------------------------------------------------------------

int main(int argc, char **argv) {
    int filename_from_config;
    int nice;
    int rc = 0;
    ofstream filelist;
    const char *filename;
    vector<string> extensions;
    string ext;
    Config settings;
    time_t nowtime;

    // Get current locale setting
    locale user_locale("");

    // Handle all options by courtesy of gengetopt
    if (cmdline_parser (argc, argv, &args_info) != 0) {
        exit(23);
    }

    // Set locale
    cout.imbue( user_locale );
    cerr.imbue( user_locale );

    // Check for debug mode
    if ( args_info.debug_arg > 0 ) {
#ifdef DEBUG
        cout << "Debug mode enabled. Writing diagnostic messages to stdout.\n" << endl;
        cout << "/======================================================" << endl;
        cout << setw(15) << left << "|Config file:" << setw(20) << args_info.config_arg << endl;
        cout << setw(15) << left << "|Debug level:" << setw(20) << args_info.debug_arg << endl;
        cout << setw(15) << left << "|Directory:"   << setw(20) << args_info.dir_arg << endl;
        cout << setw(15) << left << "|Output file:" << setw(20) << args_info.output_arg << endl;
        cout << "\\======================================================\n\n" << endl;
#else
        cout << "Sorry, this binary was compiled without debug support. Please recompile the source.\n" << endl;
#endif
    }

    // Read and parse parameters from configuration file.
    settings.readFile(args_info.config_arg);
    if ( settings.exists("extension") ) {
        // Array "extension" exists, we now extract all values and create a string vector.
        Setting& extset = settings.lookup("extension");
        int number      = extset.getLength();
#ifdef DEBUG
        if ( args_info.debug_arg > 0 ) {
            cout << "DEBUG: Extension list found in config file." << endl;
            cout << "DEBUG: Found " << number << " extensions." << endl;
        }
#endif
        // Loop through all values and copy extension strings into vector.
        for(int i=0; i<number; i++) {
            ext = (const char *)extset[i];
            extensions.push_back(ext);
#ifdef DEBUG
            if ( args_info.debug_arg > 0 ) {
                cout << "DEBUG: Extracted " << ext << endl;
            }
#endif
        }
    }
    if ( settings.lookupValue("nice",nice) ) {
#ifdef DEBUG
        if ( args_info.debug_arg > 0 ) {
            cout << "DEBUG: Nice level '" << nice << "' found in config file." << endl;
        }
#endif
    }
    if ( settings.lookupValue("output",filename) ) {
        // The filename from the command line overrides the filename in the
        // configuration file.
#ifdef DEBUG
        if ( args_info.debug_arg > 0 ) {
            cout << "DEBUG: Output filename '" << filename << "' found in config file." << endl;
        }
#endif
        filename_from_config = 1;
    }
    else {
        filename_from_config = 0;
        filename = args_info.output_arg;
    }

    // Open output file (or stream or whatever it is called in C++)
    // and deactivate C compatible streams.
    // See http://gcc.gnu.org/onlinedocs/libstdc++/27_io/howto.html#8
    std::ios::sync_with_stdio(false);
    filelist.open( filename, ios::out | ios::trunc );
    if ( filelist.is_open() ) {
        // Set locale for writing file.
        filelist.imbue( user_locale );

        // Write header
        time(&nowtime);
        filelist << "# List of files found in directory: " << args_info.dir_arg << endl;
        filelist << "# List automatically created by the filelist utility on " << ctime(&nowtime);
        // Now we walk through the directory tree.
        try {
            walkthrough( args_info.dir_arg, filelist, &extensions );
        }
        catch ( exception& err ) {
            cout << "Exception caught, aborting: " << err.what() << endl;
            rc++;
        }
        filelist.close();
    }

    // Exit
    cmdline_parser_free (&args_info);
    return(rc);
}
