/*
 * =====================================================================================
 *
 *       Filename:  helper.cpp
 *
 *    Description:  Helper functions to deal with the filesystem walk, the checks
 *                  for the correct extensions and the database interaction.
 *
 *        Version:  1.0
 *        Created:  10/26/2007 11:38:31 PM CEST
 *       Revision:  Sat Jan 12 16:13:40 CET 2008
 *       Compiler:  g++
 *
 *         Author:  R. Pfeiffer (Mr), lynx@luchs.at
 *        Company:  http://web.luchs.at/
 *
 * =====================================================================================
 */

//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

//-----------------------------------------------------------------------------
// Include section
//-----------------------------------------------------------------------------

#include <cassert>    // for assert()
#include <cerrno>     // for EILSEQ, errno
#include <cstring>    // for strlen()
#include <cwchar>     // for mbstate_t

#include <algorithm>
#include <cctype>
#include <exception>
#include <fstream>
#include <functional>
#include <iomanip>
#include <iostream>
#include <locale>
#include <map>
#include <string>
#include <sstream>
#include <vector>

#include <ctype.h>
#include <iconv.h>
#include <sqlite3.h>
#include <stdlib.h>
#include <sys/time.h>
#include <time.h>

#include <boost/algorithm/string/replace.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/spirit/core.hpp>
#include <boost/spirit/symbols/symbols.hpp>
#include <boost/spirit/utility/lists.hpp>
#include <boost/tokenizer.hpp>

#include "cmdline.h"
#include "helper.h"

using namespace std;
using namespace boost;
using namespace boost::algorithm;
using namespace boost::spirit;

namespace fs = boost::filesystem;

/*-----------------------------------------------------------------------------
 *  Externals
 *-----------------------------------------------------------------------------*/
extern gengetopt_args_info args_info;

/*-----------------------------------------------------------------------------
 *  Spirit parser for extension to filter command matching
 *-----------------------------------------------------------------------------*/
// This is the map that holds the extension/filter mappings.
map<string,string> parsed_filters;

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  AddExtension()
 *  Description:  Extracts extension from config rule.
 * =====================================================================================
 */
void AddExtension( char const* str, char const* end ) {
    string new_extension(str,end);

    parsed_filters[new_extension] = "true";
    return;
};

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  convert_utf8_to_utf32()
 *  Description:  Converts a buffer containing UTF-8 encoded text to UTF-32 encoded
 *                text (suitable for wchar_t strings).
 * =====================================================================================
 */
bool convert_utf8_to_utf32( char **source, size_t source_size, char **target, size_t target_size ) {
    iconv_t iconv_handle;
    size_t converted_chars;
    size_t ssize;
    size_t tsize;
    bool rc = false;

    if ( target_size < source_size ) {
        return(rc);
    }
    iconv_handle = iconv_open( "UCS-4", "UTF-8" );
    if ( iconv_handle != ((iconv_t)(-1)) ) {
        ssize = source_size;
        tsize = target_size;
        converted_chars = iconv( iconv_handle, source, &ssize, target, &tsize );
        if ( converted_chars != ((size_t)(-1)) ) {
            rc = true;
        }
        else {
            rc = false;
        }
        iconv_close( iconv_handle );
    }
    else {
        rc = false;
    }
    return(rc);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  convert_utf32_to_utf8()
 *  Description:  Converts a buffer containing UTF-32 encoded text to UTF-8 encoded
 *                text (suitable for wchar_t strings).
 * =====================================================================================
 */
bool convert_utf32_to_utf8( char **source, size_t source_size, char **target, size_t target_size ) {
    iconv_t iconv_handle;
    size_t converted_chars;
    size_t ssize;
    size_t tsize;
    bool rc = false;

    if ( target_size < source_size ) {
        return(rc);
    }
    iconv_handle = iconv_open( "UTF-8", "UCS-4" );
    if ( iconv_handle != ((iconv_t)(-1)) ) {
        ssize = source_size;
        tsize = target_size;
        converted_chars = iconv( iconv_handle, source, &ssize, target, &tsize );
        if ( converted_chars != ((size_t)(-1)) ) {
            rc = true;
        }
        else {
            rc = false;
        }
        iconv_close( iconv_handle );
    }
    else {
        rc = false;
    }
    return(rc);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  my_mbstowcs()
 *  Description:  Does the same as the C function mbstowcs(), but it uses the codecvt
 *                for of the C++ STL conversion.
 * =====================================================================================
 * The function was taken from:
 * http://svn.apache.org/repos/asf/incubator/stdcxx/trunk/examples/manual/mbsrtowcs.cpp
 */
std::size_t
my_mbstowcs (std::mbstate_t *pstate,
             wchar_t        *dst,
             const char     *src,
             std::size_t     size)
{
    const std::locale global;

    typedef std::codecvt<wchar_t, char, std::mbstate_t> CodeCvt;

    // retrieve the codecvt facet from the global locale
    const CodeCvt &cvt = std::use_facet<CodeCvt>(global);

    // use a small local buffer when dst is null and ignore size
    wchar_t buf [4];
    if (0 == dst) {
        dst  = buf;
        size = sizeof buf / sizeof *buf;
    }

    // set up pointers into the source sequence
    const char*       from      = src;
    const char* const from_end  = from + std::strlen (from);
    const char*       from_next = from;

    // set up pointers into the destination sequence
    wchar_t*       to     = dst;
    wchar_t* const to_end = to + size;
    wchar_t*       to_next;

    // number of non-NUL wide characters stored in destination buffer
    std::size_t nconv = 0;

    // use a local state when pstate is null (i.e., emulate mbstowcs)
    std::mbstate_t state = std::mbstate_t ();
    if (0 == pstate)
        pstate = &state;

    for ( ; from_next != from_end && to != to_end;
          from = from_next, to = dst == buf ? dst : to_next) {

        // convert a (sub)sequence of the source buffer into
        // the destination buffer
        const std::codecvt_base::result res =
            cvt.in (*pstate,
                    from, from_end, from_next,
                    to, to_end, to_next);

        // verify the consistency of the xxx_next pointers
        assert (from <= from_next && from_next <= from_end);
        assert (to <= to_next && to_next <= to_end);

        // process conversion result
        switch (res) {

        case std::codecvt_base::error:
            // conversion error
            errno = EILSEQ;
            return std::size_t (-1);

        case std::codecvt_base::noconv:
            // only codecvt<T, T> (i.e., facets where intern_type and
            // extern_type are identical) is allowed to return noconv
            // treat this case as an error even though it indicates
            // a bad (incorrectly implemented) codecvt facet
            return std::size_t (-1);

        case std::codecvt_base::partial:
            // partial conversion (incomplete character or not enough
            // room in destination buffer to convert the entire source
            // sequence)
            if (dst != buf || std::size_t (to_next - to) < size) {
                errno = EILSEQ;
                return std::size_t (-1);
            }

            nconv += to_next - to;
            break;

        case std::codecvt_base::ok:
            // complete conversion of an initial subsequence (but not
            // necessarily all) of the source buffer
            nconv += to_next - to;

            if (dst == buf && from_next == from_end)
                return nconv;

            break;
        }
    }

    return nconv;
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  widen_string()
 *  Description:  Converts a narrow string to a wide string. The encoding of the
 *                string is not changed.
 * =====================================================================================
 */
wstring widen_string( const string& str, const locale& loc ) {
    wostringstream wstm;
    wstm.imbue(loc);

    const ctype<wchar_t>& ctfacet = use_facet< ctype<wchar_t> >(wstm.getloc() );
    for( size_t i=0 ; i<str.size() ; ++i ) {
        wstm << ctfacet.widen( str[i] );
    }
    return wstm.str();
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  narrow_wstring()
 *  Description:  Converts a wide string to a narrow string. The encoding of the
 *                string is not changed (no conversion is done).
 * =====================================================================================
 */
string narrow_wstring( const wstring& str, const locale& loc ) {
    ostringstream stm;
    stm.imbue(loc);

    const ctype<char>& ctfacet = use_facet< ctype<char> >( stm.getloc() );
    for( size_t i=0 ; i<str.size() ; ++i ) {
        stm << ctfacet.narrow( str[i], 0 );
    }
    return stm.str();
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  wstring_to_multibyte()
 *  Description:  Converts a wide string to a narrow string with multibytes. The code
 *                tries to convert the characters according to the locale setting.
 * =====================================================================================
 */
string wstring_to_multibyte( const wstring& str, const locale& loc ) {
    ostringstream stm;
    stm.imbue(loc);
    string mb_buffer;

    mb_buffer.resize(10);
    for( size_t i=0 ; i<str.size() ; ++i ) {
        if ( wctomb( (char *)(mb_buffer.c_str()), str[i] ) > 0 ) {
            stm << mb_buffer;
        }
    }
    return stm.str();
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  hexdump()
 *  Description:  Prints a hexdump of a given memory region.
 * =====================================================================================
 */
#ifdef DEBUG
void print_hexdump( void *pAddressIn, unsigned long lSize )
{
 char szBuf[100];
 long lIndent = 1;
 long lOutLen, lIndex, lIndex2, lOutLen2;
 long lRelPos;
 struct { char *pData; unsigned long lSize; } buf;
 unsigned char *pTmp,ucTmp;
 unsigned char *pAddress = (unsigned char *)pAddressIn;

   buf.pData   = (char *)pAddress;
   buf.lSize   = lSize;

   while (buf.lSize > 0)
   {
      pTmp     = (unsigned char *)buf.pData;
      lOutLen  = (int)buf.lSize;
      if (lOutLen > 16)
          lOutLen = 16;

      // create a 64-character formatted output line:
      sprintf(szBuf, " >                            "
                     "                      "
                     "    %08lX", pTmp-pAddress);
      lOutLen2 = lOutLen;

      for(lIndex = 1+lIndent, lIndex2 = 53-15+lIndent, lRelPos = 0;
          lOutLen2;
          lOutLen2--, lIndex += 2, lIndex2++
         )
      {
         ucTmp = *pTmp++;

         sprintf(szBuf + lIndex, "%02X ", (unsigned short)ucTmp);
         if(!isprint(ucTmp))  ucTmp = '.'; // nonprintable char
         szBuf[lIndex2] = ucTmp;

         if (!(++lRelPos & 3))     // extra blank after 4 bytes
         {  lIndex++; szBuf[lIndex+2] = ' '; }
      }

      if (!(lRelPos & 3)) lIndex--;

      szBuf[lIndex  ]   = '<';
      szBuf[lIndex+1]   = ' ';

      printf("%s\n", szBuf);

      buf.pData   += lOutLen;
      buf.lSize   -= lOutLen;
   }

   return;
}
#endif

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  escape_xml_chars()
 *  Description:  Replaces "dangerous" characters in a string with "harmless" ones. Uses
 *                the boost string algorithm library.
 * =====================================================================================
 */
string escape_xml_chars( string xml_data ) {
    // Escape &
    replace_all(xml_data,"&","&amp;");
    // Escape <
    replace_all(xml_data,"<","&lt;");
    // Escape >
    replace_all(xml_data,">","&gt;");
    // Escape "
    replace_all(xml_data,"\"","&quot;");
    // Escape '
    replace_all(xml_data,"\'","&apos;");
    // Escape FF (replace it by white space)
    replace_all(xml_data,"\x0c"," ");

    return(xml_data);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  stringify()
 *  Description:  A quick hack to convert a number to a narrow string.
 * =====================================================================================
 */
std::string stringify( time_t x ) {
    std::ostringstream o;
    string converted;

    if (!(o << x)) {
        converted = "invalid";
    }
    else {
        converted = o.str();
    }

    return converted;
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  stringify_w()
 *  Description:  A quick hack to convert a number into a wide string.
 * =====================================================================================
 */
std::wstring stringify_w( time_t x ) {
    std::wostringstream o;
    wstring converted;

    if (!(o << x)) {
        converted = L"invalid";
    }
    else {
        converted = o.str();
    }

    return converted;
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  wstring_to_time_t()
 *  Description:  A quick hack to convert a time_t into a wide string.
 * =====================================================================================
 */
time_t wstring_to_time_t( const wchar_t *inputstring ) {
    wstringstream convertstream(inputstring);
    time_t number;

    convertstream >> number;

    return(number);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  AddFilter
 *  Description:  Extracts filter command from config file.
 * =====================================================================================
 */
void AddFilter( char const* str, char const* end ) {
    string new_filtercmd(str,end);
    map<string,string>::iterator iter;
    int true_count = 0;

    // Since the extension is parsed before the filter we should only have one map entry
    // with the default value of "true" in the whole map. This loop looks for this entry
    // and replaces "true" with the parsed filter command string.
    for(iter = parsed_filters.begin(); iter != parsed_filters.end(); iter++) {
        if ( iter->second == "true" ) {
            parsed_filters[iter->first] = new_filtercmd;
            true_count++;
        }
    }
    if ( true_count != 1 ) {
#ifdef DEBUG
        if ( args_info.debug_arg > 0 ) {
            cout << "DEBUG: Mismatched filter command " << new_filtercmd << " found in configuration file." << endl;
        }
#else
        cout << "ERROR: Mismatched filter command " << new_filtercmd << " found in configuration file." << endl;
#endif
    }

    return;
};

/*-----------------------------------------------------------------------------
 *  Configuration file grammar for extensions/filter commands
 *-----------------------------------------------------------------------------*/
struct filter_grammar : public grammar<filter_grammar>
{
    template <typename ScannerT>
    struct definition
    {
        definition(filter_grammar const& self)
        {
            char_filter_start    = alnum_p;
            char_filter_middle   = alnum_p | blank_p | ch_p('_') | ch_p('-') | ch_p('>') | ch_p('<') | ch_p('$') | ch_p('!') | ch_p('|');
            char_extension_start = alnum_p | ch_p('_');
            char_extension_middle= alnum_p | ch_p('_') | ch_p('|');
            char_start_comment   = ch_p('#') | ch_p(';') | str_p("//");
            extension            = char_extension_start >> * char_extension_middle;
            filter               = char_filter_start >> * char_filter_middle;
            blanks_p             = * blank_p;
            empty_line           = blanks_p >> eol_p;
            comment_line         = blanks_p >> char_start_comment >> * print_p >> eol_p;
            filter_line          = blanks_p >> extension [ &AddExtension ] >> blanks_p >>
                                   ch_p('=') >> blanks_p >>
                                   ch_p('(') >> filter [ &AddFilter ] >> ch_p(')') >> eol_p;
            lines                = empty_line | comment_line | filter_line;
            filter_config        = lexeme_d [ * lines ];
        }

        rule<ScannerT> char_filter_start;
        rule<ScannerT> char_filter_middle;
        rule<ScannerT> char_extension_start;
        rule<ScannerT> char_extension_middle;
        rule<ScannerT> char_start_comment;
        rule<ScannerT> extension;
        rule<ScannerT> filter;
        rule<ScannerT> blanks_p;
        rule<ScannerT> empty_line;
        rule<ScannerT> comment_line;
        rule<ScannerT> filter_line;
        rule<ScannerT> lines;
        rule<ScannerT> filter_config;

        rule<ScannerT> const& start() const { return filter_config; }
    };
};

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  create_random_string()
 *  Description:  Creates a random string of a certain length with printable characters.
 * =====================================================================================
 */
string create_random_string( size_t length ) {
    const string letters = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789";
    size_t n;
    struct timeval time;
    string random;

    gettimeofday( &time, NULL );
    srand( time.tv_usec );
    for ( n=1; n<=length; n++ ) {
        random.append( letters, rand() % letters.size(), 1 );
    }

    return(random);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  create_filter_command()
 *  Description:  Creates filter shell command by replacing $OUT with the temporary
 *                filename. $IN will be replaced by the exec_command() function because
 *                of possible white spaces in the file name.
 * =====================================================================================
 */
string create_filter_command( string tempname, string filter ) {
    string command;
    string::size_type pos;

    // The minimum length of a filter template is 9.
    // perl -e '$a="a \$IN \$OUT"; print $a," ",length($a),"\n";'
    if ( filter.length() > 9 ) {
        command = filter;
        pos = command.find("$OUT");
        if ( pos != string::npos ) {
            command.replace(pos,4,tempname);
        }
    }
    else {
        command = "false";
    }

    return(command);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  get_extension_from_filename()
 *  Description:  Analyses a filename and returns the extension. The extension is
 *                determined by searching for "." and extracting the part of the string
 *                after the last occurrence of a ".". The extension is converted to
 *                lower case.
 * =====================================================================================
 */
string get_extension_from_filename( string filename ) {
    string extension;
    size_t ext_pos = 0;
    size_t ext_length = 0;

    ext_pos = filename.rfind( filename, 0 );
    if ( ext_pos != string::npos ) {
        ext_length = filename.length() - ext_pos;
        extension  = filename.substr( ext_pos, ext_length );
    }
    else {
        // If we don't find a valid extension we return simply "." (mainly
        // because we don't wish to indicate an error condition with a 
        // fictional extension string that may turn into a legitimite some day).
        extension = ".";
    }
    return(extension);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  exec_filter()
 *  Description:  Executes filter by use of exec(), evaluates the return condition and
 *                does basic error handling.
 * =====================================================================================
 */
int exec_filter( string command, string file ) {
    const char *command_path;
    char **argv;
    vector<string> tokens;
    string::size_type pos;
    int i;
    int n;
    int rc = -1;

    typedef boost::tokenizer<boost::char_separator<char> > tokenizer;
    boost::char_separator<char> sep(" ");

    // Check if command and file name are present.
    // The filter command "false" is a code for "we don't have any filter command".
    if ( (command.length() < 1) || (file.length() < 1) || (command.compare("false") == 0) ) {
        return(rc);
    }

    // Tokenise command string.
    tokenizer tok(command,sep);
    n = 0;
    // Enclose file name with " in case any white spaces are present.
    file = "\"" + file + "\"";
    // Start loop, isolate tokens and replace $IN with file name.
    for( tokenizer::iterator tok_iter=tok.begin(); tok_iter!=tok.end(); ++tok_iter) {
        pos = tok_iter->find("$IN");
        if ( pos != string::npos ) {
            tokens.push_back(file);
        }
        else {
            tokens.push_back(*tok_iter);
        }
        n++;
    }

    argv = (char **)calloc( n+1, sizeof(char *) );
    if ( argv == NULL ) {
        cerr << "ERROR: Can't allocate memory for filter command execution (array with arguments)!" << endl;
        return(rc);
    }

    command_path = tokens[0].c_str();
    for( i=1; (i<n); i++ ) {
        argv[i] = (char *)tokens[i].c_str();
    }
#ifdef DEBUG
    if ( args_info.debug_arg > 30 ) {
        cout << "DEBUG: Turned command '" << command << "' into " << n << " tokens." << endl;
        if ( args_info.debug_arg > 20 ) {
            cout << "DEBUG: " << command_path << endl;
            for( i=1; (i<n); i++ ) {
                cout << "DEBUG: " << i << ". " << argv[i] << endl;
            }
        }
    }
#endif

    // rc = execv( command_path, argv );
    if ( command.length() > 0 ) {
        pos = command.find("$IN");
        if ( pos != string::npos ) {
            command.replace(pos,3,file);
        }
#ifdef DEBUG
        if ( args_info.debug_arg > 15 ) {
            cout << "DEBUG: system() executes '" << command << "'." << endl;
        }
#endif
        rc = system( command.c_str() );
    }

    free((void *)argv);
    return(rc);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  parse_extfilter_cfg()
 *  Description:  Parses extension filter configuration and fills a vector with all
 *                "known" extensions and their matching filter command(s).
 * =====================================================================================
 */
bool parse_extfilter_cfg( char *filebuffer ) {
    // Return code.
    bool rc;
    // The parser,
    filter_grammar filter_rules;
    // The result.
    parse_info<> parse_result;

    // Call parser.
    parse_result = parse( filebuffer, filter_rules );
    rc = parse_result.full;

    if ( !rc ) {
        cout << "ERROR: Cannot parse configuration, stopped at > " << parse_result.stop << " < " << endl;
    }

    return(rc);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  walkthrough()
 *  Description:  Walks through a given directory and creates a list with files. All
 *                files must have a valid extension given in a string vector. Files not
 *                having one of these extensions won't be listed.
 *        Notes:  This code is capable of dealing with filenames of different locales.
 *                Though we use narrow strings and no wide strings everything should be
 *                just fine as long as the filename is null-terminated. Make sure the
 *                locale set in your environment matches the encoding of the filename.
 *                The code cannot deal with mixed encodings encountered during the 
 *                tree walk (no conversions are done).
 * =====================================================================================
 */

bool walkthrough( char *directory, ofstream &logfile, vector<string> *extensions ) {
    fs::path dir_path( directory, fs::native );
    string basename;
    string basename_lc;
    string ext;
    string ext_lc;
    size_t position;
    bool rc = true;

    // We leave out anything that starts with a ".".
#ifdef DEBUG
    if ( args_info.debug_arg > 10 ) {
        cout << "DEBUG: Entering '" << directory << "'" << endl;
    }
#endif
    // Get the basename / leaf of the current entry.
    basename = dir_path.leaf();
    if ( basename[0] == '.' ) {
#ifdef DEBUG
        if ( args_info.debug_arg > 0 ) {
            cout << "DEBUG: Leaving out '" << directory << "'" << endl;
        }
#endif
        return(true);
    }
    // Check if we start with a directory.
    if ( fs::is_directory(directory) ) {
        // Iterate through whole directory.
        fs::directory_iterator end_itr;
        for( fs::directory_iterator itr( dir_path );
             itr != end_itr;
             ++itr ) {
            // If the entry is a directory we have to list this as well.
            if ( fs::is_directory(*itr) ) {
#ifdef DEBUG
                if ( args_info.debug_arg > 0 ) {
                    string file = itr->native_file_string().c_str();
                    cout << "DEBUG: Found directory '" << file << "'" << endl;
                }
#endif
                rc = walkthrough( (char *)(itr->native_file_string().c_str()), logfile, extensions );
                // We don't return here because we only return when having processed
                // a file (otherweise the recursion breaks at the first directory).
            }
            else {
                // We found a file.
                basename = itr->leaf();
#ifdef DEBUG
                if ( args_info.debug_arg > 0 ) {
                    cout << "DEBUG: Found file '" << basename << "'" << endl;
                }
#endif
                // Check for extension(s); we need to do this strictly in lower case, but we
                // need to remember the original file name.
                basename_lc = basename;
                transform(basename_lc.begin(), basename_lc.end(), basename_lc.begin(), (int(*)(int))std::tolower);
                for(unsigned int i=0; i < extensions->size(); i++) {
                    ext    = extensions->at(i);
                    ext_lc = ext;
                    transform(ext_lc.begin(), ext_lc.end(), ext_lc.begin(), (int(*)(int))std::tolower);
                    position = basename_lc.rfind(ext_lc);
                    if ( position != string::npos ) {
                        // We found an extension and compare.
                        if ( position == (basename_lc.length()-ext_lc.length()) ) {
                            // We have a match and write it to the log file.
                            // We use the full path information when writing to
                            // the log.
                            logfile << itr->native_file_string().c_str() << endl;
#ifdef DEBUG
                            if ( args_info.debug_arg > 10 ) {
                                cout << "DEBUG: Wrote file '" << itr->native_file_string().c_str() << "' to log." << endl;
                            }
#endif
                        }
                    }
                } // of for() loop for extensions
            } // of file processing
        } // of iterator for() loop
    } // of directory check
    else {
        rc = false;
    }

    return(rc);
}

