#ifndef INCLUDE_HELPER_H
#define INCLUDE_HELPER_H
/*
 * =====================================================================================
 * 
 *       Filename:  helper.h
 * 
 *    Description:  Prototypes and definitions for the helper.cc code.
 * 
 *        Version:  1.0
 *        Created:  10/28/2007 11:43:33 PM CET
 *       Revision:  Sat Jan 12 16:28:12 CET 2008
 *       Compiler:  g++
 * 
 *         Author:  R. Pfeiffer (Mr), lynx@luchs.at
 *        Company:  http://web.luchs.at/
 * 
 * =====================================================================================
 */

#include <cctype>
#include <iostream>
#include <functional>
#include <locale>
#include <string>
#include <vector>

#include <sqlite3.h>
#include <string.h>
#include <time.h>

using namespace std;

/*-----------------------------------------------------------------------------
 *  Definition & limits
 *-----------------------------------------------------------------------------*/
#define FILENAME_MAX_LENGTH     65536

/*-----------------------------------------------------------------------------
 *  Command line options
 *-----------------------------------------------------------------------------*/
extern gengetopt_args_info args_info;

/*-----------------------------------------------------------------------------
 *  Data structure for extension filter
 *-----------------------------------------------------------------------------*/
struct extfilter {
    string extension;
    string filtercmd;
};

/*
 * === TEMPLATE ========================================================================
 *        Name: 
 * Description: Allow string conversion from std::string to std::wstring.
 *       Usage: Declare something like "Widen<wchar_t> to_wstring;" and then use
 *              to_wstring().
 *      Source: http://www.kharchi.eu/wiki/doku.php?id=cpp:std:string
 * =====================================================================================
 */

template<class E,
class T = std::char_traits<E>,
class A = std::allocator<E> >
 
class Widen : public std::unary_function<
   const std::string&, std::basic_string<E, T, A> >
{
     std::locale loc_;
     const std::ctype<E>* pCType_;
 
     // No copy-constructor, no assignment operator...
     Widen(const Widen&);
     Widen& operator= (const Widen&);
 
 public:
     // Constructor...
     Widen(const std::locale& loc = std::locale()) : loc_(loc)
     {
#if defined(_MSC_VER) && (_MSC_VER < 1300) // VC++ 6.0...
         using namespace std;
         pCType_ = &_USE(loc, ctype<E> );
#else
         pCType_ = &std::use_facet<std::ctype<E> >(loc);
#endif
     }
 
     // Conversion...
     std::basic_string<E, T, A> operator() (const std::string& str) const
     {
         typename std::basic_string<E, T, A>::size_type srcLen =
             str.length();
         const char* pSrcBeg = str.c_str();
         std::vector<E> tmp(srcLen);
 
         pCType_->widen(pSrcBeg, pSrcBeg + srcLen, &tmp[0]);
         return std::basic_string<E, T, A>(&tmp[0], srcLen);
     }
};

/*-----------------------------------------------------------------------------
 *  SQLite interaction
 *-----------------------------------------------------------------------------*/
/* Interaction with SQLite is done by a class. */

class SQLite3 {
    private:
        sqlite3 *db;
        char *sql_error;
        const char *stmt_error;
        char **result;
        int rc;
        int number_rows,number_columns;
        int db_is_open;

        // Data structures for prepared statement.
        sqlite3_stmt *stmt_insert;
        sqlite3_stmt *stmt_select;
        sqlite3_stmt *stmt_update;

    public:
        vector<string> vcol_head;
        vector<string> vdata;
        time_t mtime;

        // Class constructor
        SQLite3 (string tablename="default.db"): sql_error(0), rc(0), db_is_open(0) {
            rc = sqlite3_open_v2(tablename.c_str(), &db, SQLITE_OPEN_READWRITE | SQLITE_OPEN_CREATE, NULL );
            if( rc ) {
                cout << "ERROR: Can't open SQLite database: " << sqlite3_errmsg(db) << endl;
                sqlite3_close(db);
            }
            sqlite3_extended_result_codes( db, 1 );
            db_is_open = 1;
        }

        // Class destructor
        ~SQLite3() {
//            int code;

//            code = sqlite3_finalize(stmt_insert);
//            if ( code != SQLITE_OK ) {
//                cerr << "ERROR: sqlite3_finalize() failed with " << sqlite3_errmsg(db) << endl;
//            }
//            code = sqlite3_finalize(stmt_select);
//            if ( code != SQLITE_OK ) {
//                cerr << "ERROR: sqlite3_finalize() failed with " << sqlite3_errmsg(db) << endl;
//            }
//            code = sqlite3_finalize(stmt_update);
//            if ( code != SQLITE_OK ) {
//                cerr << "ERROR: sqlite3_finalize() failed with " << sqlite3_errmsg(db) << endl;
//            }
            sqlite3_close(db);
            db_is_open = 0;
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  prepare_insert()
         *  Description:  Prepares SQL statement for INSERT operation.
         * =====================================================================================
         */
        bool prepare_insert() {
            int code;
            bool rc = true;
            const char *sql_insert = "INSERT INTO fileaccess VALUES (?,?)";

            code = sqlite3_prepare_v2( db, sql_insert, -1, &stmt_insert, &stmt_error );
            if ( code != SQLITE_OK ) {
                rc = false;
                cerr << "ERROR: Preparing INSERT failed. Ignoring everything after " << stmt_error << endl;
            }
#ifdef DEBUG
            if ( args_info.debug_arg > 0 ) {
                cout << "DEBUG: Prepared statement: " << sql_insert << endl;
            }
#endif

            return(rc);
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  prepare_select()
         *  Description:  Prepares SQL statement for SELECT operation.
         * =====================================================================================
         */
        bool prepare_select() {
            int code;
            bool rc = true;
            const char *sql_select = "SELECT mtime FROM fileaccess WHERE filename=?";

            code = sqlite3_prepare_v2( db, sql_select, -1, &stmt_select, &stmt_error );
            if ( code != SQLITE_OK ) {
                rc = false;
                cerr << "ERROR: Preparing SELECT failed. Ignoring everything after " << stmt_error << endl;
            }
#ifdef DEBUG
            if ( args_info.debug_arg > 0 ) {
                cout << "DEBUG: Prepared statement: " << sql_select << endl;
            }
#endif
            return(rc);
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  prepare_select()
         *  Description:  Prepares SQL statement for SELECT operation.
         * =====================================================================================
         */
        bool prepare_update() {
            int code;
            bool rc = true;
            const char *sql_update = "UPDATE fileaccess SET mtime=? WHERE filename=?";

            code = sqlite3_prepare_v2( db, sql_update, -1, &stmt_update, &stmt_error );
            if ( code != SQLITE_OK ) {
                rc = false;
                cerr << "ERROR: Preparing UPDATE failed. Ignoring everything after " << stmt_error << endl;
            }
#ifdef DEBUG
            if ( args_info.debug_arg > 0 ) {
                cout << "DEBUG: Prepared statement: " << sql_update << endl;
            }
#endif
            return(rc);
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  insert_filename()
         *  Description:  Inserts filename and mtime into table.
         * =====================================================================================
         */
        bool insert_filename( string filename, time_t mtime ) {
            int code;
            bool rc = true;

            code = sqlite3_bind_text( stmt_insert, 1, filename.c_str(), filename.length(), SQLITE_STATIC );
            if ( code != SQLITE_OK ) {
                rc = false;
                cerr << "ERROR: Can't insert filename " << filename << " into SQL statement. ("
                     << sqlite3_errmsg(db) << ")" << endl;
            }
            code = sqlite3_bind_int64( stmt_insert, 2, (sqlite3_int64)(mtime) );
            if ( code != SQLITE_OK ) {
                rc = false;
                cerr << "ERROR: Can't insert mtime " << mtime << " into SQL INSERT statement. ("
                     << sqlite3_errmsg(db) << ")" << endl;
            }
            if ( rc == true ) {
                code = sqlite3_step( stmt_insert );
                if ( (code != SQLITE_OK) && (code != SQLITE_DONE) ) {
                    rc = false;
                    cerr << "ERROR: Can't execute SQL (" << sqlite3_errmsg(db) << ", " << code << ")" << endl;
                }
#ifdef DEBUG
                else {
                    if ( args_info.debug_arg > 20 ) {
                        cout << "DEBUG: Inserted filename with time stamp " << mtime << "." << endl;
                    }
                }
#endif
            }
            sqlite3_finalize( stmt_insert );
            return(rc);
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  select_mtime()
         *  Description:  Retrieves mtime corresponding to filename.
         * =====================================================================================
         */
        time_t select_mtime( string filename ) {
            int code;
            time_t mtime = 23;
            bool rc = true;

            code = sqlite3_bind_text( stmt_select, 1, filename.c_str(), strlen(filename.c_str()), SQLITE_STATIC );
            if ( code != SQLITE_OK ) {
                rc = false;
                cerr << "ERROR: Can't insert filename " << filename << " into SQL SELECT statement. ("
                     << sqlite3_errmsg(db) << ", code " << code << ")" << endl;
            }
            if ( rc == true ) {
                code = sqlite3_step( stmt_select );
                if ( (code != SQLITE_OK) && (code != SQLITE_DONE) && (code != SQLITE_ROW) ) {
                    rc = false;
                    cerr << "ERROR: Can't execute SELECT SQL (" << sqlite3_errmsg(db) << ", " << code << ")" << endl;
                }
                else {
                    mtime = (time_t)sqlite3_column_int64( stmt_select, 0 );
                }
            }
            if ( rc == false ) {
                mtime = 0;
            }
            sqlite3_finalize( stmt_select );
            return(mtime);
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  update_mtime()
         *  Description:  Updates mtime for a given filename.
         * =====================================================================================
         */
        bool update_mtime( string filename, time_t mtime ) {
            int code;
            bool rc = true;

#ifdef DEBUG
            if ( args_info.debug_arg > 10 ) {
                cout << "DEBUG: Updating entry for " << filename.c_str() << " with length " 
                     << strlen(filename.c_str()) << "." << endl;
            }
#endif
            code = sqlite3_bind_text( stmt_update, 2, filename.c_str(), -1, SQLITE_STATIC );
            if ( code != SQLITE_OK ) {
                rc = false;
                cerr << "ERROR: Can't insert filename " << filename << " into SQL UPDATE statement. ("
                     << sqlite3_errmsg(db) << ")" << endl;
            }
            code = sqlite3_bind_int64( stmt_update, 1, (sqlite3_int64)(mtime) );
            if ( code != SQLITE_OK ) {
                cerr << "ERROR: Can't insert mtime " << mtime << " into SQL statement. ("
                     << sqlite3_errmsg(db) << ")" << endl;
            }
            if ( rc == true ) {
                code = sqlite3_step( stmt_update );
                if ( (code != SQLITE_OK) && (code != SQLITE_DONE) ) {
                    rc = false;
                    cerr << "ERROR: Can't execute SQL (" << sqlite3_errmsg(db) << ")" << endl;
                }
//                sqlite3_reset( stmt_select );
            }
            sqlite3_finalize( stmt_update );
            return(rc);
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  get_error_msg()
         *  Description:  Returns the current error message from the SQLite engine.
         * =====================================================================================
         */
        const char *get_error_msg() {
            return( sqlite3_errmsg(db) );
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  get_error_code()
         *  Description:  Returns the current error code from the SQLite engine.
         * =====================================================================================
         */
        int get_error_code() {
            return( sqlite3_errcode(db) );
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  is_open()
         *  Description:  Checks if database is open or not and returns the state as boolean.
         * =====================================================================================
         */
        bool is_open() {
            if ( db_is_open == 1 ) {
                return(true);
            }
            else {
                return(false);
            }
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  execute()
         *  Description:  Sends SQL command to SQLite engine.
         * =====================================================================================
         */
        int execute( string sql ) {
            rc = sqlite3_exec( db, sql.c_str(), NULL, NULL, &sql_error );
            if ( rc != SQLITE_OK ) {
                cout << "ERROR: SQL command didn't succeed ( " << sqlite3_errmsg(db) << ", " 
                     << sql_error << " )" << endl;
                sqlite3_free((void *)sql_error);
            }
            return(rc);
        }

        /* 
         * ===  FUNCTION  ======================================================================
         *         Name:  get_table()
         *  Description:  Sends SQL string to SQLite engine.
         * =====================================================================================
         */
        int get_table( string sql ) {
            rc = sqlite3_get_table( db, sql.c_str(), &result, &number_rows, &number_columns, &sql_error );

            if (vcol_head.size()<0) {
                vcol_head.clear();
            }
            if (vdata.size()<0) {
                vdata.clear();
            }


            if ( rc == SQLITE_OK ) {
               for(int i=0; i < number_columns; ++i) {
                   vcol_head.push_back(result[i]);
               }
               for(int i=0; i < number_columns*number_rows; ++i) {
                   vdata.push_back(result[number_columns+i]);
               }
            }
            else {
                cerr << "ERROR: get_table() failed ( " << sqlite3_errmsg(db) << " )" << endl;
            }
            sqlite3_free_table(result);
            return rc;
        }
};

//----------------------------------------------------------------------
// Functions
//----------------------------------------------------------------------
extern bool convert_utf8_to_utf32( char **source, size_t source_size, char **target, size_t target_size );
extern bool convert_utf32_to_utf8( char **source, size_t source_size, char **target, size_t target_size );
extern string narrow_wstring( const wstring& str, const locale& loc );
extern wstring widen_string( const string& str, const locale& loc );
extern string wstring_to_multibyte( const wstring& str, const locale& loc );

extern string create_filter_command( string tempname, string filter );
extern string create_random_string( size_t length );
extern string get_extension_from_filename( string filename );
extern int exec_filter( string command, string file );
extern std::size_t my_mbstowcs ( std::mbstate_t *pstate, wchar_t *dst, const char *src, std::size_t size );
extern bool parse_extfilter_cfg( char *filebuffer );
extern void print_hexdump( void *pAddressIn, unsigned long lSize );

extern string escape_xml_chars( string xml_data );
extern std::string stringify( time_t x );
extern std::wstring stringify_w( time_t x );
extern time_t wstring_to_time_t( const wchar_t *inputstring );
extern bool walkthrough( char *directory, ofstream &logfile, vector<string> *extensions );

#endif
