/*
 * =====================================================================================
 *
 *       Filename:  indexer.cc
 *
 *    Description:  Main indexer that reads a list of files to be indexed, checks their
 *                  modification time and feeds them to the indexing process for indexing
 *                  or reindexing. A data conversion to the text format is done before
 *                  calling the index function.
 *
 *        Version:  1.0
 *        Created:  10/28/2007 11:40:57 PM CET
 *       Revision:  12/15/2007 14:48:01 PM CET
 *       Compiler:  g++
 *
 *         Author:  R. Pfeiffer (Mr), lynx@luchs.at
 *        Company:  http://web.luchs.at/
 *
 * =====================================================================================
 */

//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

//-----------------------------------------------------------------------------
// Strategy
//-----------------------------------------------------------------------------
// We already have a list of files to be indexed. This means that we won't do
// any recursive walks through the filesystem (at least not big ones).
//
// - Parse config file.
//   - Extract list of extensions and matching filters.
// - Open/create database.
// - Read list of files and create a string vector.
// - Optionally sort the entries by directory.
// - Loop through whole vector.
//   - Stat file.
//   - Compared mtime with value in database.
//   - Insert/update mtime in database 
//     - Determine file type by extension.
//     - Lookup suitable filter for data conversion.
//       - Convert file to text and write it to temporary space.
//     - Feed text data to index.
// - Done.

//-----------------------------------------------------------------------------
// Text filter configuration
//-----------------------------------------------------------------------------
//
// Filter methods must be given by the following syntax:
// extension = (command -i $IN -o $OUT)
//
// This format is used internally as follows:
// string = (string)
//
// The leftmost string indicates the extension. When a conversion needs
// to be done, the rightmost string is used and $IN/$OUT are replaced by
// the according file names. An empty string "()" indicates that no
// conversion should be done.
//
// The filters are defined in the configuration file. The code uses no
// internal defaults.
//

// Useful resources:
// http://beans.seartipy.com/2006/05/10/boost-filesystem-library-writing-portable-c-programs-to-acess-the-filesystem/
// http://www.daniweb.com/forums/thread74944.html
// http://freshmeat.net/articles/view/1428/
//

//-----------------------------------------------------------------------------
// Include section
//-----------------------------------------------------------------------------

#include <algorithm>
#include <iostream>
#include <fstream>
#include <iomanip>
#include <string>
#include <exception>
#include <locale>
#include <vector>

#include <time.h>
#include <sqlite3.h>

#include <libconfig.h++>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/filesystem/path.hpp>

#include "indexer_cmdline.h"
#include "helper.h"
#include "indexer.h"
#include "indexer_functions.h"

using namespace libconfig;
using namespace std;

namespace fs = boost::filesystem;

//-----------------------------------------------------------------------------
// Definitions
//-----------------------------------------------------------------------------
#define CREATE_ACCESS_TABLE "CREATE TABLE IF NOT EXISTS fileaccess ( filename TEXT PRIMARY KEY, mtime INT(8) )"
#define BEGIN_TRANSACTION   "BEGIN"
#define COMMIT_TRANSACTION  "COMMIT"

#define RETURN_CODE_FAIL    13
#define RETURN_CODE_OK      0 

//-----------------------------------------------------------------------------
// Globals
//-----------------------------------------------------------------------------  

// Filter commands mapped to extension. This is a global storage that holds the
// information parsed from the configuration file.
extern map<string,string> parsed_filters;

// Structure with command line options
gengetopt_args_info args_info;

// Temporary directory with '/' at the end
string global_temp_dir;

//----------------------------------------------------------------------
// Main section
//----------------------------------------------------------------------

int main(int argc, char **argv) {
    vector<string> allfiles;
    ifstream config_file;
    ifstream fileslist;
    string config_raw;
    string line;
    time_t last_run;
    time_t nowtime;
    bool new_index = false;
    int rc = RETURN_CODE_OK;
    
    // Handle all options by courtesy of gengetopt
    if (cmdline_parser (argc, argv, &args_info) != 0) {
        exit(23);
    }

    // Check temporary directory.
    global_temp_dir.assign(args_info.tmp_arg);
    if ( global_temp_dir[global_temp_dir.size()-1] != '/' ) {
        global_temp_dir.append("/");
    }

    // Check for debug mode
    if ( args_info.debug_arg > 0 ) {
#ifdef DEBUG
        cout << "Debug mode enabled. Writing diagnostic messages to stdout.\n" << endl;
        cout << "/======================================================" << endl;
        cout << setw(20) << left << "|Config file:"     << setw(30) << args_info.config_arg << endl;
        cout << setw(20) << left << "|Debug level:"     << setw(30) << args_info.debug_arg << endl;
        cout << setw(20) << left << "|Database:"        << setw(30) << args_info.database_arg << endl;
        cout << setw(20) << left << "|File list:"       << setw(30) << args_info.list_arg << endl;
        cout << setw(20) << left << "|Index directory:" << setw(30) << args_info.index_arg << endl;
        cout << setw(20) << left << "|New index flag:"  << setw(30) << args_info.new_arg << endl;
        cout << setw(20) << left << "|Period (s):"      << setw(30) << args_info.period_arg << endl;
        cout << setw(20) << left << "|Temporary dir:"   << setw(30) << global_temp_dir << endl;
        cout << setw(20) << left << "|Size of wchar_t:" << setw(30) << sizeof(wchar_t) << endl;
        cout << "\\======================================================\n\n" << endl;
#else
        cout << "Sorry, this binary was compiled without debug support. Please recompile the source.\n" << endl;
#endif
    }

    // ---------------------------------------------------------------------------
    // Check command line parameters and verify required resources and input.

    // Check if index directory is really a directory.
    fs::path index_path( args_info.index_arg );
    if ( ! fs::is_directory( index_path ) ) {
        cerr << "ERROR: Index directory is no directory (it may even not exist)." << endl;
        exit(23);
    }

    // Get time.
    time(&nowtime);
    // Compute date of changes to look for.
    last_run = nowtime - args_info.period_arg;
    if ( last_run < 0 ) {
        last_run = nowtime - 86400;
    }

    // Extension filters
    fs::path configuration(args_info.config_arg);
    if ( fs::exists(configuration) && (fs::is_directory(configuration) == false) ) {
        // All is fine, read the configuration and parse the content filters.
        config_file.open(args_info.config_arg,ifstream::in);
        if ( config_file.is_open() ) {
            while ( !config_file.eof() ) {
                getline( config_file, line );
                if ( line.length() > 1 ) {
                    config_raw += line + "\n";
                }
            }
            config_file.close();
#ifdef DEBUG
            if ( args_info.debug_arg > 15 ) {
                cout << "DEBUG: Configuration file yielded:" << endl;
                cout << "DEBUG: ---------------------------" << endl;
                cout << config_raw << endl;
                cout << "DEBUG: ---------------------------" << endl;
            }
#endif
            if ( ! parse_extfilter_cfg( (char *)config_raw.c_str() ) ) {
                cerr << "ERROR: Cannot parse filter rules!" << endl;
                rc = RETURN_CODE_FAIL;
            }
#ifdef DEBUG
            if ( args_info.debug_arg > 15 ) {
                cout << "DEBUG: --- Map ---------------" << endl;
                map<string,string>::iterator iter;
                for(iter = parsed_filters.begin(); iter != parsed_filters.end(); iter++) {
                    cout << iter->first << " maps to " << iter->second << endl;
                }
                cout << "DEBUG: -----------------------" << endl;
            }
#endif
        }
        else {
            cerr << "ERROR: Cannot open file " << configuration.leaf() << "!" << endl;
            rc = RETURN_CODE_FAIL;
        }
    }
    else {
        cerr << "ERROR: Configuration file " << configuration.leaf() << " does not exist." << endl;
        rc = RETURN_CODE_FAIL;
    }

    // Check if SQLite database exists. Create a new one if necessary.
    SQLite3 sql_db( args_info.database_arg );
    if ( !sql_db.is_open() ) {
        cerr << "ERROR: Can't open database! Reason: " << sql_db.get_error_msg() << endl;
        rc = RETURN_CODE_FAIL;
    }
    else {
        sql_db.execute(CREATE_ACCESS_TABLE);
    }

    // List of files to parse.
    fs::path filelist(args_info.list_arg);
    if ( fs::exists(filelist) && (fs::is_directory(filelist) == false) && (rc==0) ) {
        // We got the list, we now need to read all entries and create a vector
        fileslist.open(args_info.list_arg,ifstream::in);
        if ( fileslist.is_open() ) {
            while ( !fileslist.eof() ) {
                getline( fileslist, line );
                if ( (line.length() > 1) && (line[0]!='#') ) {
                    allfiles.push_back(line);
                }
            }
            fileslist.close();
            sort(allfiles.begin(),allfiles.end());
#ifdef DEBUG
            if ( args_info.debug_arg > 0 ) {
                cout << "DEBUG: Read " << allfiles.size() << " files from file list." << endl;
            }
#endif
        }
        else {
            cout << "ERROR: Cannot open file " << filelist.leaf() << "!" << endl;
            rc = RETURN_CODE_FAIL;
        }
    }

    // Now we have everything we need and we can begin processing all files.
    // We do that in one big transaction. There might be better ways, but until we find
    // ways for improvement, we do it this way.
    if ( rc == RETURN_CODE_OK ) {
#ifdef DEBUG
        if ( args_info.debug_arg > 0 ) {
            cout << "DEBUG: Starting SQL transaction and calling index_files()." << endl;
            cout << endl;
        }
#endif
        new_index = ( args_info.new_arg!=0 ? true : false );
        sql_db.execute(BEGIN_TRANSACTION);
        rc = index_files( allfiles, &sql_db, last_run, new_index );
        if ( rc == RETURN_CODE_OK ) {
            // Commit transaction provided we encountered no critical errors.
            sql_db.execute(COMMIT_TRANSACTION);
        }
        else {
            cerr << "ERROR: Commit of transaction failed! Reason: " << sql_db.get_error_msg() << endl;
            rc = RETURN_CODE_FAIL;
        }
    }

    return(rc);
}

