/*
 * =====================================================================================
 *
 *       Filename:  indexer_functions.cc
 *
 *    Description:  Code for indexing functions, converters and database access.
 *
 *        Version:  1.0
 *        Created:  11/02/2007 09:06:44 PM CET
 *       Revision:  Sat Jan 12 16:24:24 CET 2008
 *       Compiler:  g++
 *
 *         Author:  R. Pfeiffer (Mr), lynx@luchs.at
 *        Company:  http://web.luchs.at/
 *
 * =====================================================================================
 */

//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include <cctype>
#include <cstdlib>
#include <cstring>
#include <fstream>
#include <iostream>
#include <string>
#include <vector>

#include <errno.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <time.h>

#include <CLucene.h>
#include <CLucene/util/Reader.h>
#include <CLucene/index/IndexWriter.h>
#include <CLucene/document/Document.h>
#include <CLucene/document/Field.h>
#include <CLucene/util/stringreader.h>
#include <CLucene/analysis/Analyzers.h>

#include <boost/filesystem/operations.hpp>
#include <boost/filesystem/convenience.hpp>
#include <boost/filesystem/path.hpp>

#include <sqlite3.h>

#include <libconfig.h++>

#include "indexer_cmdline.h"
#include "helper.h"

using namespace std;

namespace fs = boost::filesystem;

// CLucene has a lot of namespace branches
using namespace lucene::index;
using namespace lucene::analysis;
using namespace lucene::util;
using namespace lucene::store;
using namespace lucene::document;
using namespace lucene::search;
using namespace lucene::queryParser;

/*-----------------------------------------------------------------------------
 *  Externals
 *-----------------------------------------------------------------------------*/
extern gengetopt_args_info args_info;
extern string global_temp_dir;
extern map<string,string> parsed_filters;

/*-----------------------------------------------------------------------------
 *  Definitions
 *-----------------------------------------------------------------------------*/

#define POPEN_PIPE_LENGTH              512
#define TEMPORARY_FILENAME_LENGTH       32

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  get_file_type()
 *  Description:  This is a wrapper for the file command. It calls file, give the name
 *                of the file as argument and captures the standard output in a string.
 *                We use popen() for this (which is a hack, but so far meta information
 *                is optional at this stage).
 * =====================================================================================
 */
std::wstring get_file_type( string filename ) {
    FILE *stream;
    string command;
    string o;
    wstring output;
    size_t output_length;
    size_t c_pos;

    command = "file -b -p \"" + filename +"\"";
    stream  = popen( command.c_str(), "r" );
    if ( stream != NULL ) {
        o.resize(POPEN_PIPE_LENGTH);
        fread( (void *)(&o[0]), sizeof(char), POPEN_PIPE_LENGTH, stream );
        pclose( stream );
        output_length = ::mbstowcs( NULL, o.c_str(), o.length() );
        output.resize(output_length);
        ::mbstowcs( &output[0], o.c_str(), o.length() );
        // Remove linefeed and newline from the end of the string.
        c_pos = output.find(L"\n");
        if ( c_pos != string::npos ) {
            output.erase( c_pos, 1 );
        }
        c_pos = output.find(L"\r");
        if ( c_pos != string::npos ) {
            output.erase( c_pos, 1 );
        }
    }
    else {
        output = L"No information.";
    }

    return(output);
}

/* 
 * ===  FUNCTION  ======================================================================
 *         Name:  index_files()
 *  Description:  Walk through the file list, stat every file, compare mtime with mtime
 *                in database and feed file to converter/indexer if necessary.
 * =====================================================================================
 */
int index_files( vector<string> allfiles, SQLite3 *db, time_t last_run, bool new_index ) {
    // Used for file processing
    vector<string>::iterator file_itr;
    string file_extension;
    string file_temporary;
    size_t file_extpos;
    string filter_command;
    fs::path* tempfile;

    // Wide string conversion template
    Widen<wchar_t> to_wstring;

    // File information
    struct stat fileinfo;
    time_t mtime;
    time_t new_mtime;

    wchar_t *file_content;
    std::wstring filename_type;
    std::wstring filename_wide;

    // Used for filename conversion
    std::mbstate_t state = std::mbstate_t();

    ifstream::pos_type file_size;
    wifstream file_stream;
    bool file_has_content;
    bool index_file;
    bool new_file;

    // Fields we want to put into the index
    lucene::document::Field *field_filename;
    lucene::document::Field *field_file_content;
    lucene::document::Field *field_mtime;
    lucene::document::Field *field_type;

    // CLucene document container
    lucene::document::Document *file_document;

    // Return code of this function (rc), stat() call (sc) and exec() (ec) call
    int ec = 0;
    int rc = 0;
    int sc = 0;

    // This is the analyser we use.
    WhitespaceAnalyzer analyser;

    // Initialise CLucene index writer
    IndexWriter::IndexWriter index_repository( args_info.index_arg, &analyser, new_index, true );

    // ---------------------------------------------------------------------------------------
    // Processing loop starts here
    // ---------------------------------------------------------------------------------------

    // We take the time of the last run and look for all mtimes that are smaller.
    // Loop through all files.
    for( file_itr = allfiles.begin(); file_itr != allfiles.end(); file_itr++ ) {
        // Per default we mark the file as having no content.
        file_has_content = false;
        sc = stat( file_itr->c_str(), &fileinfo );
        // We only process files that can be stat()ed.
        if ( sc == 0 ) {
            db->prepare_select();
            mtime = db->select_mtime( *file_itr );
#ifdef DEBUG
            if ( args_info.debug_arg > 10 ) {
                cout << "DEBUG: Retrieved mtime " << mtime << " from database." << endl;
            }
#endif
            // Check if we have a new file or not. A mtime of 0 means we have a new
            // file.
            if ( mtime == 0 ) {
                new_file = true;
            }
            else {
                new_file = false;
            }
            // We use the file's new mtime from now on since we know if the file is
            // new or not. We update the mtime in the database if the file's 
            // mtime is bigger.
            new_mtime = fileinfo.st_mtime;
            if ( (fileinfo.st_mtime > last_run) || (new_file == true) ) {
                // File is newer than the period chosen, index it. In order to do this
                // we need to determine the extension, check if we have a suitable
                // filter command, execute the filter if present and shove everything
                // into the index.
                // CLucene's document representation consists of key/value pairs. We 
                // are free to create as many pairs as we want. Two is the minimum.
                //
                // filename => name of the file
                // content  => (filtered) content of the file
                //
                // Write file with mtime into database.
                if ( new_file == true ) {
                    db->prepare_insert();
                    db->insert_filename( *file_itr, new_mtime );
                    index_file = true;
                }
                else {
                    // Only update if necessary.
                    if ( new_mtime > mtime ) {
                        db->prepare_update();
                        db->update_mtime( *file_itr, new_mtime );
                        index_file = true;
                    }
                    else {
                        index_file = false;
                    }
                }

                // Convert data if necessary:
                // 1. Determine extension of file
                // 2. Determine if conversion is in the parsed configuration
                //    a. Create shell command
                //    b. Execute shell command
                //    c. Add converted content to index
                // 3. Add filename to index (without content)

                // Determine extension of file
                file_extpos = file_itr->rfind(".");
                if ( file_extpos != string::npos ) {
                    file_extension = file_itr->substr( file_extpos+1 );
#ifdef DEBUG
                    if ( args_info.debug_arg > 15 ) {
                        cout << "DEBUG: Found extension '" << file_extension << "'." << endl;
                        cout << "DEBUG: Filter lookup in table yields: " << parsed_filters[file_extension] << endl;
                    }
#endif
                    // Look if we have a suitable filter command
                    // The value true is used as a placeholder when filling the string map. A value different
                    // from true means we have a filter command that can be excuted.
                    if ( parsed_filters[file_extension] != "true" ) {
                        // Create random name for temporary file and make sure the path of
                        // the temporary disk space is in front.
                        file_temporary = create_random_string( TEMPORARY_FILENAME_LENGTH );
                        file_temporary.insert(0,global_temp_dir);
                        filter_command = create_filter_command( file_temporary, parsed_filters[file_extension] );
#ifdef DEBUG
                        if ( args_info.debug_arg > 15 ) {
                            cout << "DEBUG: Using filter command '" << filter_command << "' for extraction." << endl;
                        }
#endif
                        // Execute filter
                        // The exec_filter() function uses similar return values as the system() call.
                        if ( exec_filter( filter_command, *file_itr ) == -1 ) {
                            ec = 1;
                            file_has_content = false;
                            cerr << "ERROR: Filter failed!" << endl;
                        }
                        else {
                            // Mark return code as successful.
                            // Nevertheless mark that file has no content until
                            // the content buffer is really filled.
                            ec = 0;
                            file_has_content = false;

#ifdef DEBUG
                            if ( args_info.debug_arg > 15 ) {
                                cout << "DEBUG: About to read file into memory buffer." << endl;
                            }
#endif

                            // Create Boost path object from temporary file
                            tempfile = new fs::path (file_temporary);

                            // Create field with output from filter.
                            // We open with a I/O stream class and add its content (i.e. the file)
                            // into the allocated field object.
                            file_stream.open( file_temporary.c_str(), ios_base::in | ios_base::binary | ios_base::ate );

                            if ( !file_stream.is_open() ) {
                                cerr << "ERROR: Can't open file " << file_temporary << " for reading!" << endl;
                                file_has_content = false;
                            }
                            else {
                                // Read file into memory buffer.
                                //
                                // One word about encoding: The shell utilities usually produce a stream
                                // encoded in UTF-8. The CLucene methods use the TCHAR type for compatibility
                                // reasons. TCHAR is really a wchar_t and is used to represent text. So
                                // either we need to convert between dynamic character length strings and wide
                                // character strings (GCC uses 4 byte, some other compilers use 2) or we
                                // use wide I/O stream functions that deliver data in wchar_t format.
                                // If you don't watch out for this, then your CLucene index will contain
                                // unintelligible characters.
                                // 
                                // http://www.ioplex.com/~miallen/libmba/dl/docs/ref/text_details.html
                                // http://www.boost.org/libs/serialization/doc/codecvt.html
                                // http://gcc.gnu.org/onlinedocs/libstdc++/22_locale/codecvt.html
                                // http://www.icu-project.org/

                                file_size = file_stream.tellg();
#ifdef DEBUG
                                if ( args_info.debug_arg > 15 ) {
                                    cout << "DEBUG: Got a file size of " << file_size << " bytes." << endl;
                                }
#endif
                                // We only process files with content.
                                if ( file_size > 0 ) {
                                    file_content = new wchar_t[file_size];
                                    file_stream.seekg( 0, ios_base::beg );
                                    file_stream.read( file_content, (streamsize)file_size );
                                    file_has_content = true;
                                    // We add one byte to make sure we have room for null termination at the end
                                    // (we might not need it, but C habits never die).
                                    file_size += sizeof(char);

#ifdef DEBUG
                                    if ( args_info.debug_arg > 15 ) {
                                        cout << "DEBUG: UTF-8 buffer size is " << sizeof(char) * file_size << " bytes. ";
                                        cout << "UCS-4 buffer size is " << sizeof(wchar_t) * file_size << " bytes." << endl;
                                    }
#endif
                                    field_file_content = new Field::Field( (const wchar_t*) L"content",
                                                                           (const wchar_t*)file_content,
                                                                           true, true, true, false );
#ifdef DEBUG
                                    // Just print out what we just read from the file (along with some characters
                                    // for debugging UTF-8 output).
                                    if ( args_info.debug_arg > 30 ) {
                                        cout << "DEBUG: Read content -- René écris plus de ÛTF-8 ☥ ---" << endl;
                                        print_hexdump( (void *)file_content, (unsigned long)(sizeof(wchar_t)*file_size) );
                                        cout << "DEBUG: ----------------------------------------------" << endl;
                                    }
#endif
                                } // of file size check
                                else {
                                    file_has_content = false;
                                    // if ( file_size == -1 ) {
                                    //    cout << "ERROR: Can't read file! (" << << ")" << endl;
                                    // }
                                }

                                // Close stream
                                file_stream.close();

                            } // of read file into memory buffer
                        } // of exec_filter()

                    } // of check for extension filter
                }
                // Create field for storing the filename. We need to be careful to use the correct
                // encoding for the filename.
                //filename_wide_length = ::mbstowcs( NULL, file_itr->c_str(), file_itr->length() );
                //filename_wide.resize(filename_wide_length);
                //::mbstowcs( &filename_wide[0], file_itr->c_str(), file_itr->length() );
                //if ( my_mbstowcs( &state, &filename_wide[0], file_itr->c_str(), filename_wide_length ) == filename_wide_length ) {
                //    filename_wide[filename_wide_length] = L'\0';
                //}
                //else {
                //    filename_wide = L"invalid filename";
                //}
                filename_wide = to_wstring(*file_itr);
#ifdef DEBUG
                if ( args_info.debug_arg > 15 ) {
                    wcout << "DEBUG: Converted wide file name is '" << filename_wide << "'." << endl;
                }
#endif
                field_filename = new Field::Field( (const wchar_t*) L"filename", 
                                                   filename_wide.c_str(),
                                                   true, true, true, false );

                // Call file command and get some meta information.
                filename_type = get_file_type( *file_itr );
                field_type  = new Field::Field( (const wchar_t*) L"type",
                                                filename_type.c_str(),
                                                true, true, true, false );

                // Create field for storing mtime.
                field_mtime = new Field::Field( (const wchar_t*) L"mtime", 
                                                (stringify_w(new_mtime)).c_str(),
                                                true, true, true, false );
#ifdef DEBUG
                if ( args_info.debug_arg > 20 ) {
                    cout << "DEBUG: Created mtime field with string '" << stringify(new_mtime) << "'." << endl;
                }
#endif

                // Create Lucene document for adding to index directory.
                file_document = new lucene::document::Document;
                // Add fields to document object.
                if ( file_has_content ) {
                    file_document->add( *field_file_content );
                }
                file_document->add( *field_filename );
                file_document->add( *field_mtime );
                file_document->add( *field_type );
                index_repository.addDocument( file_document, &analyser );

#ifdef DEBUG
                if ( args_info.debug_arg > 26 ) {
                    cout << "DEBUG: --- Created document ---------------------------------" << endl;
                    cout << "DEBUG: Filename " << field_filename->toString() << endl;
                    cout << "DEBUG: mtime " << field_mtime->toString() << endl;
                    wprintf( L"DEBUG: printf -> %s\nDEBUG: printf -> %s\n", field_filename->toString(), 
                             field_mtime->toString() );
                    if ( file_has_content ) {
                        cout << "DEBUG: Document " << file_document->toString() << endl;
                        wprintf( L"DEBUG: printf -> %s\n", file_document->toString() );
                    }
                    cout << "DEBUG: ------------------------------------------------------" << endl;
                }
#endif

                // Clean up temporary file (if it exists).
                if ( fs::exists(*tempfile) && (ec != 1) ) {
                    fs::remove(*tempfile);
                }

                // Free objects used for file processing.
                if ( tempfile && (ec != 1) ) {
                    delete tempfile;
                }
                if ( field_filename ) {
                    delete field_filename;
                }
                if ( field_type ) {
                    delete field_type;
                }
                if ( file_content ) {
                    // We double check if the file's content was really processed.
                    if ( file_has_content ) {
                        delete[] file_content;
                    }
                }
                if ( field_file_content ) {
                    // We double check if the file's content was really processed.
                    if ( file_has_content ) {
                        delete field_file_content;
                    }
                }
            } // of mtime check
#ifdef DEBUG
            else {
                if ( args_info.debug_arg > 10 ) {
                    cout << "DEBUG: Skipping file " << *file_itr << "." << endl;
                }
            }
#endif
        } // of stat() call
        else {
            cerr << "ERROR: Can't stat() file " << *file_itr << " (" << strerror(errno) << ")" << endl;
        }
    } // of iterator loop

    // Optimising the index should be done after there were changes to the index.
    index_repository.optimize();

    // Close index.
    index_repository.close();

    return(rc);
}


